@@ -1,3 +1,7 @@
+// Copyright 2015 The Go Authors. All rights reserved.
+// Use of this source code is governed by a BSD-style
+// license that can be found in the LICENSE file.
+
 package mips
 
 var cnames0 = []string{
@@ -22,6 +26,7 @@ var cnames0 = []string{
 	"LACON",
 	"LECON",
 	"DACON",
+	"STCON",
 	"SBRA",
 	"LBRA",
 	"SAUTO",
@@ -33,6 +38,7 @@ var cnames0 = []string{
 	"LOREG",
 	"GOK",
 	"ADDR",
+	"TLS",
 	"TEXTSIZE",
 	"NCLASS",
 }
